@@ -4,17 +4,27 @@
  *	apply function to files
  */
 
-#include <fcntl.h>
 #include "fd.h"
+#include "time.h"
+#include "realpath.h"
+#include "parse.h"
 #include "func.h"
 #include "kanji.h"
 
-#define	MAXTIMESTR	8
-#define	ATTRWIDTH	10
+#define	MAXTIMESTR		8
+#define	ATTRWIDTH		10
 #if	MSDOS
-#define	DIRENTSIZ(s)	DOSDIRENT
+#define	DIRENTSIZ(s)		DOSDIRENT
 #else
-#define	DIRENTSIZ(s)	(sizeof(struct dirent) - DNAMESIZE + strlen(s) + 1)
+#define	DIRENTSIZ(s)		(sizeof(struct dirent) \
+				- DNAMESIZE + strlen(s) + 1)
+#endif
+#ifdef	_NOEXTRACOPY
+#define	MAXCOPYITEM		4
+#define	FLAG_SAMEDIR		0
+#else
+#define	MAXCOPYITEM		5
+#define	FLAG_SAMEDIR		8
 #endif
 
 typedef struct _attrib_t {
@@ -41,50 +51,39 @@ extern int lcmdline;
 extern int maxcmdline;
 extern int mark;
 #ifdef	HAVEFLAGS
-extern u_long fflaglist[];
+extern CONST u_long fflaglist[];
 #endif
 #if	!defined (_USEDOSCOPY) && !defined (_NOEXTRACOPY)
 extern int inheritcopy;
 #endif
 
-#ifndef	EISDIR
-#define	EISDIR	EACCES
-#endif
-
-#ifdef	_NOEXTRACOPY
-#define	MAXCOPYITEM	4
-#define	FLAG_SAMEDIR	0
-#else
-#define	MAXCOPYITEM	5
-#define	FLAG_SAMEDIR	8
-#endif
-
-static int NEAR issamedir __P_((char *, char *));
+static int NEAR issamedir __P_((CONST char *, CONST char *));
 static int NEAR islowerdir __P_((VOID_A));
-static char *NEAR getdestdir __P_((char *, char *));
-static int NEAR getdestpath __P_((char *, char *, struct stat *));
-static int NEAR getcopypolicy __P_((char *));
-static int NEAR getremovepolicy __P_((char *, int));
-static int NEAR checkdupl __P_((char *, char *, struct stat *, struct stat *));
-static int NEAR checkrmv __P_((char *, int));
+static char *NEAR getdestdir __P_((CONST char *, CONST char *));
+static int NEAR getdestpath __P_((CONST char *, char *, struct stat *));
+static int NEAR getcopypolicy __P_((CONST char *));
+static int NEAR getremovepolicy __P_((CONST char *, int));
+static int NEAR checkdupl __P_((CONST char *, char *,
+		struct stat *, struct stat *));
+static int NEAR checkrmv __P_((CONST char *, int));
 #ifndef	_NOEXTRACOPY
-static int NEAR isxdev __P_((char *, struct stat *));
+static int NEAR isxdev __P_((CONST char *, struct stat *));
 static VOID NEAR addcopysize __P_((off_t));
 static VOID NEAR _showprogress __P_((int));
-static int countcopysize __P_((char *));
-static int countmovesize __P_((char *));
-static int NEAR prepareprogress __P_((int, int (*)(char *)));
+static int countcopysize __P_((CONST char *));
+static int countmovesize __P_((CONST char *));
+static int NEAR prepareprogress __P_((int, int (*)(CONST char *)));
 #endif
 static int NEAR preparecopy __P_((int, int));
 static int NEAR preparemove __P_((int, int));
-static int safecopy __P_((char *));
-static int safemove __P_((char *));
-static int cpdir __P_((char *));
-static int touchdir __P_((char *));
+static int safecopy __P_((CONST char *));
+static int safemove __P_((CONST char *));
+static int cpdir __P_((CONST char *));
+static int touchdir __P_((CONST char *));
 #ifndef	_NOEXTRACOPY
-static int mvdir1 __P_((char *));
-static int mvdir2 __P_((char *));
-static int countremovesize __P_((char *));
+static int mvdir1 __P_((CONST char *));
+static int mvdir2 __P_((CONST char *));
+static int countremovesize __P_((CONST char *));
 #endif
 static VOID NEAR showmode __P_((attrib_t *, int, int));
 static VOID NEAR showattr __P_((namelist *, attrib_t *, int));
@@ -93,11 +92,12 @@ static int NEAR inputuid __P_((attrib_t *, int));
 static int NEAR inputgid __P_((attrib_t *, int));
 #endif
 static char **NEAR getdirtree __P_((char *, char **, int *, int));
-static int NEAR _applydir __P_((char *, int (*)(char *),
-		int (*)(char *), int (*)(char *), int, char *, int));
-static int forcecpfile __P_((char *));
-static int forcecpdir __P_((char *));
-static int forcetouchdir __P_((char *));
+static int NEAR _applydir __P_((CONST char *, int (*)(CONST char *),
+		int (*)(CONST char *), int (*)(CONST char *),
+		int, CONST char *, int));
+static int forcecpfile __P_((CONST char *));
+static int forcecpdir __P_((CONST char *));
+static int forcetouchdir __P_((CONST char *));
 
 char *destpath = NULL;
 #ifndef	_NOEXTRACOPY
@@ -128,10 +128,10 @@ static time_t destatime = (time_t)-1;
 #ifndef	_NOEXTRACOPY
 static char *forwardpath = NULL;
 #endif
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_PSEUDOPATH
 static int destdrive = -1;
 #endif
-#if	!defined (_NOEXTRACOPY) && !defined (_NODOSDRIVE)
+#if	!defined (_NOEXTRACOPY) && defined (DEP_DOSDRIVE)
 static int forwarddrive = -1;
 #endif
 #ifndef	_NOEXTRACOPY
@@ -145,93 +145,61 @@ static off_t maxcopysize = (off_t)0;
 
 
 static int NEAR issamedir(path, org)
-char *path, *org;
+CONST char *path, *org;
 {
-	char *cwd;
-	int i;
+	char path2[MAXPATHLEN], org2[MAXPATHLEN];
 
-	cwd = getwd2();
-	if (!org) org = cwd;
-	else if (_chdir2(org) >= 0) {
-		org = getwd2();
-		if (_chdir2(cwd) < 0) error(cwd);
-	}
-	else {
-		free(cwd);
-		return(0);
-	}
-
-#ifdef	_USEDOSPATH
-	if (dospath(path, NULL) != dospath(org, NULL)) path = NULL;
-	else
+	org = (org) ? getrealpath(org, org2, NULL) : Xgetwd(org2);
+	if (org != org2) return(0);
+#ifdef	DEP_DOSPATH
+	if (dospath(path, NULL) != dospath(org2, NULL)) return(0);
 #endif
-	path = (_chdir2(path) >= 0) ? getwd2() : NULL;
-	if (_chdir2(cwd) < 0) error(cwd);
-	i = (path) ? !strpathcmp(org, path) : 0;
-	if (org != cwd) free(org);
-	free(cwd);
-	if (path) free(path);
+	if (getrealpath(path, path2, NULL) != path2) return(0);
+	if (strpathcmp(path2, org2)) return(0);
 
-	return(i);
+	return(1);
 }
 
 static int NEAR islowerdir(VOID_A)
 {
-#ifdef	_USEDOSEMU
+#ifdef	DEP_DOSEMU
 	char orgpath[MAXPATHLEN];
 #endif
-	char *cp, *top, *cwd, *path, *org, tmp[MAXPATHLEN];
-	int i;
+	CONST char *org;
+	char *cp, *top, path[MAXPATHLEN], buf[MAXPATHLEN];
+	int len;
 
-	cwd = getwd2();
-	path = destpath;
-	org = filelist[filepos].name;
-	if (!org) org = cwd;
-#ifdef	_USEDOSEMU
-	else org = nodospath(orgpath, org);
-#endif
-	strcpy(tmp, path);
-	top = tmp;
-#ifdef	_USEDOSPATH
-	if (dospath(path, NULL) != dospath(org, NULL)) {
-		free(cwd);
+	if (!isdir(&(filelist[filepos])) || islink(&(filelist[filepos])))
 		return(0);
-	}
-	if (_dospath(tmp)) top += 2;
+	org = filelist[filepos].name;
+#ifdef	DEP_DOSEMU
+	org = nodospath(orgpath, org);
+#endif
+	strcpy2(path, destpath);
+	top = path;
+#ifdef	DEP_DOSPATH
+	if (dospath(path, NULL) != dospath(org, NULL)) return(0);
+	if (_dospath(path)) top += 2;
 #endif
-	while (_chdir2(tmp) < 0) {
-		if ((cp = strrdelim(top, 0)) && cp > top) *cp = '\0';
-		else {
-			path = NULL;
-			break;
-		}
-	}
-	i = 0;
-	if (path) {
-		path = getwd2();
-		if (_chdir2(cwd) < 0) error(cwd);
-		if (_chdir2(org) >= 0) {
-			org = getwd2();
-			i = strlen(org);
-			if (strnpathcmp(org, path, i) || path[i] != _SC_)
-				i = 0;
-			free(org);
-		}
-		if (_chdir2(cwd) < 0) error(cwd);
-		free(path);
-	}
-	free(cwd);
 
-	return(i);
+	while ((cp = (char *)getrealpath(path, buf, NULL)) != buf) {
+		if (cp != path) return(0);
+		cp = strrdelim(top, 0);
+		if (!cp || cp <= top) return(0);
+		*cp = '\0';
+	}
+
+	if (getrealpath(org, path, NULL) != path) return(0);
+	len = strlen(path);
+	if (strnpathcmp(path, buf, len) || buf[len] != _SC_) len = 0;
+
+	return(len);
 }
 
 static char *NEAR getdestdir(mes, arg)
-char *mes, *arg;
+CONST char *mes, *arg;
 {
 	char *dir;
-#ifndef	_NODOSDRIVE
-	int drive;
-#endif
 
 	if (arg && *arg) dir = strdup2(arg);
 	else if (!(dir = inputstr(mes, 1, -1, NULL, HST_PATH))) return(NULL);
@@ -240,7 +208,7 @@ char *mes, *arg;
 		dir[0] = '.';
 		dir[1] = '\0';
 	}
-#ifdef	_USEDOSPATH
+#ifdef	DEP_DOSPATH
 	else if (_dospath(dir) && !dir[2]) {
 		dir = realloc2(dir, 4);
 		dir[2] = '.';
@@ -248,19 +216,19 @@ char *mes, *arg;
 	}
 #endif
 
-#ifndef	_NODOSDRIVE
-	destdrive = -1;
-	if ((drive = dospath3(dir)) && (destdrive = preparedrv(drive)) < 0) {
+#ifdef	DEP_PSEUDOPATH
+	if ((destdrive = preparedrv(dir, NULL, NULL)) < 0) {
 		warning(-1, dir);
-		free(dir);
+		free2(dir);
 		return(NULL);
 	}
 #endif
 	if (preparedir(dir) < 0) {
 		warning(-1, dir);
-		free(dir);
-#ifndef	_NODOSDRIVE
-		if (destdrive >= 0) shutdrv(destdrive);
+		free2(dir);
+#ifdef	DEP_DOSDRIVE
+		shutdrv(destdrive);
+		destdrive = -1;
 #endif
 		return(NULL);
 	}
@@ -269,22 +237,24 @@ char *mes, *arg;
 }
 
 static int NEAR getdestpath(file, dest, stp)
-char *file, *dest;
+CONST char *file;
+char *dest;
 struct stat *stp;
 {
-	char *cp, *tmp;
+	CONST char *cp;
+	char *tmp;
 	int n;
 
-	strcpy(dest, destpath);
+	strcpy2(dest, destpath);
 	cp = file;
 	if (destdir) {
 		for (n = 0; (tmp = strdelim(cp, 0)); n++) cp = tmp + 1;
 		for (tmp = destdir; n > 0 && (tmp = strdelim(tmp, 0)); tmp++)
 			n--;
 		if (tmp) *tmp = '\0';
-		if (*destdir) strcpy(strcatdelim(dest), destdir);
+		if (*destdir) strcpy2(strcatdelim(dest), destdir);
 	}
-	strcpy(strcatdelim(dest), cp);
+	strcpy2(strcatdelim(dest), cp);
 
 	if (Xlstat(file, stp) < 0) {
 		warning(-1, file);
@@ -295,15 +265,15 @@ struct stat *stp;
 }
 
 static int NEAR getcopypolicy(s)
-char *s;
+CONST char *s;
 {
 #ifndef	_NOEXTRACOPY
-# ifndef	_NODOSDRIVE
+# ifdef	DEP_PSEUDOPATH
 	int dupdestdrive;
 # endif
 	char *cp;
 #endif	/* !_NOEXTRACOPY */
-	char *str[MAXCOPYITEM];
+	CONST char *str[MAXCOPYITEM];
 	int n, ch, val[MAXCOPYITEM];
 
 	for (;;) {
@@ -335,18 +305,19 @@ char *s;
 			val[0] = CPP_FORWUPDATE;
 			val[1] = CPP_FORWOVERWRITE;
 			if (selectstr(&n, 2, 0, str, val) != K_CR) continue;
-# ifndef	_NODOSDRIVE
+# ifdef	DEP_PSEUDOPATH
 			dupdestdrive = destdrive;
+			destdrive = -1;
 # endif
 
 			if (!(cp = getdestdir(FRWDD_K, NULL))) continue;
-# ifndef	_NODOSDRIVE
+# ifdef	DEP_PSEUDOPATH
 			forwarddrive = destdrive;
 			destdrive = dupdestdrive;
 # endif
 			if (issamedir(cp, NULL) || issamedir(cp, destpath)) {
 				warning(EINVAL, cp);
-				free(cp);
+				free2(cp);
 				continue;
 			}
 			forwardpath = cp;
@@ -361,10 +332,10 @@ char *s;
 }
 
 static int NEAR getremovepolicy(s, pre)
-char *s;
+CONST char *s;
 int pre;
 {
-	char *str[4];
+	CONST char *str[4];
 	int i, n, ch, val[4];
 
 	Xlocate(0, L_CMDLINE);
@@ -397,18 +368,20 @@ int pre;
 }
 
 static int NEAR checkdupl(file, dest, stp1, stp2)
-char *file, *dest;
+CONST char *file;
+char *dest;
 struct stat *stp1, *stp2;
 {
 #ifndef	_NOEXTRACOPY
 	char path[MAXPATHLEN];
 #endif
+	CONST char *s;
 	char *cp, *tmp;
 	int i, n;
 
 	if (getdestpath(file, dest, stp1) < 0) return(CHK_ERROR);
 	if (Xlstat(dest, stp2) < 0) {
-		stp2 -> st_mode = S_IWRITE;
+		stp2 -> st_mode = S_IWUSR;
 		if (errno == ENOENT) return(CHK_OK);
 		warning(-1, dest);
 		return(CHK_ERROR);
@@ -427,9 +400,9 @@ struct stat *stp1, *stp2;
 
 	for (;;) {
 		if (!n || n == CPP_RENAME) {
-			cp = SAMEF_K;
-			i = strlen2(cp) - strsize("%.*s");
-			cp = asprintf3(cp, n_column - i, dest);
+			s = SAMEF_K;
+			i = strlen2(s) - strsize("%.*s");
+			cp = asprintf3(s, n_column - i, dest);
 
 #ifndef	_NOEXTRACOPY
 			copycolumn = -1;
@@ -440,7 +413,7 @@ struct stat *stp1, *stp2;
 				Xputterm(L_CLEAR);
 				Xkanjiputs(cp);
 			}
-			free(cp);
+			free2(cp);
 			if (n < 0) return(n);
 		}
 		switch (n) {
@@ -454,10 +427,10 @@ struct stat *stp1, *stp2;
 				lcmdline = L_INFO;
 				tmp = inputstr(NEWNM_K, 1, -1, NULL, -1);
 				if (!tmp) return(CHK_ERROR);
-				strcpy(getbasename(dest), tmp);
-				free(tmp);
+				strcpy2(getbasename(dest), tmp);
+				free2(tmp);
 				if (Xlstat(dest, stp2) < 0) {
-					stp2 -> st_mode = S_IWRITE;
+					stp2 -> st_mode = S_IWUSR;
 					if (errno == ENOENT) return(CHK_OK);
 					warning(-1, dest);
 					return(CHK_ERROR);
@@ -505,27 +478,37 @@ struct stat *stp1, *stp2;
 }
 
 static int NEAR checkrmv(path, mode)
-char *path;
+CONST char *path;
 int mode;
 {
-	char *cp;
-	int n, len;
-#if	MSDOS
-
-	if (Xaccess(path, mode) >= 0) return(CHK_OK);
-#else	/* !MSDOS */
+#if	!MSDOS
 	struct stat *stp, st;
 	char *tmp, dir[MAXPATHLEN];
+	int duperrno;
+#endif
+	CONST char *s;
+	char *cp;
+	int n, len;
 
-# ifndef	_NODOSDRIVE
+#if	MSDOS
+	if (Xaccess(path, mode) >= 0) return(CHK_OK);
+#else	/* !MSDOS */
+# ifdef	DEP_DOSDRIVE
 	if (dospath2(path)) {
 		if (Xaccess(path, mode) >= 0) return(CHK_OK);
 		stp = NULL;
 	}
 	else
+# endif
+# ifdef	DEP_URLPATH
+	if (urlpath(path, NULL, NULL, NULL)) {
+		if (Xaccess(path, mode) >= 0) return(CHK_OK);
+		stp = NULL;
+	}
+	else
 # endif
 	{
-		if (nodoslstat(path, &st) < 0) {
+		if (reallstat(path, &st) < 0) {
 			warning(-1, path);
 			return(CHK_ERROR);
 		}
@@ -535,7 +518,7 @@ int mode;
 		else if (tmp == path) copyrootpath(dir);
 		else strncpy2(dir, path, tmp - path);
 
-		if (nodoslstat(dir, &st) < 0) {
+		if (reallstat(dir, &st) < 0) {
 			warning(-1, dir);
 			return(CHK_ERROR);
 		}
@@ -552,15 +535,8 @@ int mode;
 
 #if	!MSDOS
 	if (stp) {
-		int duperrno;
-
 		duperrno = errno;
-# ifdef	NOUID
-		mode = logical_access(stp -> st_mode)
-# else
-		mode = logical_access(stp -> st_mode,
-			stp -> st_uid, stp -> st_gid);
-# endif
+		mode = logical_access2(stp);
 		if (!(mode & F_ISWRI)) {
 			errno = duperrno;
 			warning(-1, path);
@@ -571,14 +547,14 @@ int mode;
 
 	if (removepolicy > 0) n = removepolicy;
 	else {
-		cp = DELPM_K;
-		len = strlen2(cp) - strsize("%.*s");
-		cp = asprintf3(cp, n_column - len, path);
+		s = DELPM_K;
+		len = strlen2(s) - strsize("%.*s");
+		cp = asprintf3(s, n_column - len, path);
 #ifndef	_NOEXTRACOPY
 		copycolumn = -1;
 #endif
 		n = getremovepolicy(cp, 0);
-		free(cp);
+		free2(cp);
 	}
 	if (n > 0) n -= RMP_BIAS;
 
@@ -588,14 +564,14 @@ int mode;
 #ifndef	_NOEXTRACOPY
 /*ARGSUSED*/
 static int NEAR isxdev(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 # if	!MSDOS
 	struct stat st1, st2;
 # endif
 
-# ifdef	_USEDOSPATH
+# ifdef	DEP_DOSPATH
 	if (dospath(path, NULL) != dospath(destpath, NULL)) {
 		if (stp && Xlstat(path, stp) < 0) return(-1);
 		return(1);
@@ -624,14 +600,16 @@ off_t size;
 static VOID NEAR _showprogress(n)
 int n;
 {
-	int i, max;
+	int i, max, duperrno;
 
+	duperrno = errno;
 	max = n_column - 1;
 	Xlocate(0, copyline);
 	for (i = 0; i < n; i++) Xputch2('o');
 	for (; i < max; i++) Xputch2('.');
 	Xlocate(n, copyline);
 	Xtflush();
+	errno = duperrno;
 }
 
 VOID showprogress(size)
@@ -668,19 +646,19 @@ off_t size;
 }
 
 VOID fshowprogress(path)
-char *path;
+CONST char *path;
 {
-	char *cp;
+	CONST char *cp;
 
 	if (!(cp = strrdelim(path, 1))) cp = path;
 	showprogress(DIRENTSIZ(cp));
 }
 
 static int countcopysize(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
-	char *cp;
+	CONST char *cp;
 
 	if (!(cp = strrdelim(path, 1))) cp = path;
 	addcopysize(DIRENTSIZ(cp));
@@ -691,11 +669,11 @@ char *path;
 }
 
 static int countmovesize(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	int n;
-	char *cp;
+	CONST char *cp;
 
 	if (!(cp = strrdelim(path, 1))) cp = path;
 	addcopysize(DIRENTSIZ(cp));
@@ -707,7 +685,7 @@ char *path;
 
 static int NEAR prepareprogress(isdir, func)
 int isdir;
-int (*func)__P_((char *));
+int (*func)__P_((CONST char *));
 {
 	int ret;
 
@@ -747,7 +725,7 @@ int isdir, narg;
 		if (getcopypolicy(PRECP_K) < 0) return(-1);
 	if (prepareprogress(isdir, countcopysize)) return(-1);
 #endif	/* !_NOEXTRACOPY */
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_DOSDRIVE
 	if (dospath3(nullstr)) waitmes();
 #endif
 
@@ -771,7 +749,7 @@ int isdir, narg;
 }
 
 static int safecopy(path)
-char *path;
+CONST char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
@@ -785,7 +763,7 @@ char *path;
 }
 
 static int safemove(path)
-char *path;
+CONST char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
@@ -800,7 +778,7 @@ char *path;
 }
 
 static int cpdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
@@ -835,13 +813,13 @@ char *path;
 			break;
 	}
 
-	if (!(destmode & S_IWRITE)) {
+	if (!(destmode & S_IWUSR)) {
 		st1.st_nlink = TCH_MODE;
-		st1.st_mode = (destmode | S_IWRITE);
+		st1.st_mode = (destmode | S_IWUSR);
 		if (touchfile(dest, &st1) >= 0) destnlink |= TCH_MODE;
 	}
 
-	if (destdir) free(destdir);
+	free2(destdir);
 	destdir = &(dest[strlen(destpath)]);
 	while (*destdir == _SC_) destdir++;
 	destdir = strdup2(destdir);
@@ -851,7 +829,7 @@ char *path;
 
 /*ARGSUSED*/
 static int touchdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
@@ -875,7 +853,7 @@ char *path;
 
 #ifndef	_NOEXTRACOPY
 static int mvdir1(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	int n;
@@ -886,9 +864,9 @@ char *path;
 	}
 	if ((n = cpdir(path)) < 0) return(n);
 	if (Xlstat(path, &st) < 0) return(APL_ERROR);
-	if (!(st.st_mode & S_IWRITE)) {
+	if (!(st.st_mode & S_IWUSR)) {
 		st.st_nlink = TCH_MODE;
-		st.st_mode |= S_IWRITE;
+		st.st_mode |= S_IWUSR;
 		if (touchfile(path, &st) < 0) return(APL_ERROR);
 	}
 
@@ -896,7 +874,7 @@ char *path;
 }
 
 static int mvdir2(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
@@ -916,9 +894,9 @@ char *path;
 }
 
 static int countremovesize(path)
-char *path;
+CONST char *path;
 {
-	char *cp;
+	CONST char *cp;
 
 	if (!(cp = strrdelim(path, 1))) cp = path;
 	addcopysize(DIRENTSIZ(cp));
@@ -942,7 +920,7 @@ int isdir, narg;
 }
 
 int rmvfile(path)
-char *path;
+CONST char *path;
 {
 	int n;
 
@@ -957,7 +935,7 @@ char *path;
 }
 
 int rmvdir(path)
-char *path;
+CONST char *path;
 {
 	int n;
 
@@ -972,7 +950,7 @@ char *path;
 }
 
 int findfile(path)
-char *path;
+CONST char *path;
 {
 	if (regexp_exec(findregexp, getbasename(path), 1)) {
 		if (path[0] == '.' && path[1] == _SC_) path += 2;
@@ -989,7 +967,7 @@ char *path;
 }
 
 int finddir(path)
-char *path;
+CONST char *path;
 {
 	if (regexp_exec(findregexp, getbasename(path), 1)) {
 		if (yesno(FOUND_K)) {
@@ -1173,11 +1151,11 @@ int yy;
 	else {
 		lcmdline = yy;
 		warning(ENOENT, s);
-		free(s);
+		free2(s);
 		return(-1);
 	}
 
-	free(s);
+	free2(s);
 	if (uid != attr -> uid) {
 		attr -> uid = uid;
 		attr -> nlink |= TCH_UID;
@@ -1209,11 +1187,11 @@ int yy;
 	else {
 		lcmdline = yy;
 		warning(ENOENT, s);
-		free(s);
+		free2(s);
 		return(-1);
 	}
 
-	free(s);
+	free2(s);
 	if (gid != attr -> gid) {
 		attr -> gid = gid;
 		attr -> nlink |= TCH_GID;
@@ -1232,15 +1210,16 @@ int flag;
 #ifdef	HAVEFLAGS
 	char buf[WMODE + 1];
 #endif
-	struct tm *tm;
 	attrib_t attr;
+	struct tm *tm;
+	time_t t;
 	u_int mask;
 	int ch, x, y, xx, yy, ymin, ymax, dupwin_x, dupwin_y, excl;
 
 	dupwin_x = win_x;
 	dupwin_y = win_y;
 	subwindow = 1;
-	Xgetkey(-1, 0);
+	Xgetkey(-1, 0, 0);
 
 	yy = filetop(win);
 	while (yy + WMODELINE + 5 > n_line - 1) yy--;
@@ -1294,7 +1273,7 @@ int flag;
 		else
 #endif
 		mask = (1 << (8 - x));
-		switch (ch = Xgetkey(1, 0)) {
+		switch (ch = Xgetkey(1, 0, 0)) {
 			case K_UP:
 #if	!defined (_NOEXTRAATTR) && !defined (NOUID)
 				if (x > ATTRWIDTH) {
@@ -1526,7 +1505,7 @@ int flag;
 	win_x = dupwin_x;
 	win_y = dupwin_y;
 	subwindow = 0;
-	Xgetkey(-1, 0);
+	Xgetkey(-1, 0, 0);
 
 	if (ch == K_ESC) return(0);
 
@@ -1547,8 +1526,9 @@ int flag;
 
 	if (tm -> tm_mon < 0 || tm -> tm_mon > 11
 	|| tm -> tm_mday < 1 || tm -> tm_mday > 31
-	|| tm -> tm_hour > 23 || tm -> tm_min > 59 || tm -> tm_sec > 59)
+	|| tm -> tm_hour > 23 || tm -> tm_min > 59 || tm -> tm_sec > 60)
 		return(-1);
+	if ((t = timelocal2(tm)) == (time_t)-1) return(-1);
 
 	mask = (TCH_CHANGE | TCH_MASK | TCH_MODEEXE);
 	switch (excl) {
@@ -1564,15 +1544,16 @@ int flag;
 			break;
 #endif
 		default:
-			if (attrmode != namep -> st_mode) mask |= TCH_MODE;
+			if (attr.mode != (namep -> st_mode & ~S_IFMT))
+				mask |= TCH_MODE;
 #ifdef	HAVEFLAGS
-			if (attrflags != namep -> st_flags) mask |= TCH_FLAGS;
+			if (attr.flags != namep -> st_flags) mask |= TCH_FLAGS;
 #endif
 #if	!defined (_NOEXTRAATTR) && !defined (NOUID)
-			if (attruid != namep -> st_uid) mask |= TCH_UID;
-			if (attrgid != namep -> st_gid) mask |= TCH_GID;
+			if (attr.uid != namep -> st_uid) mask |= TCH_UID;
+			if (attr.gid != namep -> st_gid) mask |= TCH_GID;
 #endif
-			if (attrtime != namep -> st_mtim) mask |= TCH_MTIME;
+			if (t != namep -> st_mtim) mask |= TCH_MTIME;
 			break;
 	}
 	attrnlink = (attr.nlink & mask);
@@ -1587,13 +1568,13 @@ int flag;
 	attrgid = attr.gid;
 # endif
 #endif	/* !_NOEXTRAATTR */
-	attrtime = timelocal2(tm);
+	attrtime = t;
 
 	return(1);
 }
 
 int setattr(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 
@@ -1616,10 +1597,10 @@ char *path;
 }
 
 int applyfile(func, endmes)
-int (*func)__P_((char *));
-char *endmes;
+int (*func)__P_((CONST char *));
+CONST char *endmes;
 {
-#ifdef	_USEDOSEMU
+#ifdef	DEP_DOSEMU
 	char path[MAXPATHLEN];
 #endif
 	int i, ret, old, dupfilepos;
@@ -1687,7 +1668,7 @@ int *maxp, depth;
 	cp = strcatdelim(dir);
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
-		strcpy(cp, dp -> d_name);
+		strcpy2(cp, dp -> d_name);
 
 		if (Xlstat(dir, &st) < 0 || !s_isdir(&st)) continue;
 		dirlist = b_realloc(dirlist, *maxp, char *);
@@ -1704,12 +1685,12 @@ int *maxp, depth;
 }
 
 static int NEAR _applydir(dir, funcf, funcd1, funcd2, order, endmes, verbose)
-char *dir;
-int (*funcf)__P_((char *));
-int (*funcd1)__P_((char *));
-int (*funcd2)__P_((char *));
+CONST char *dir;
+int (*funcf)__P_((CONST char *));
+int (*funcd1)__P_((CONST char *));
+int (*funcd2)__P_((CONST char *));
 int order;
-char *endmes;
+CONST char *endmes;
 int verbose;
 {
 	DIR *dirp;
@@ -1717,7 +1698,8 @@ int verbose;
 	struct stat st;
 	time_t dupmtime, dupatime;
 	u_short dupmode;
-	char *cp, *fname, path[MAXPATHLEN], **dirlist;
+	CONST char *cp;
+	char *fname, path[MAXPATHLEN], **dirlist;
 	int ret, ndir, max, dupnlink;
 
 	if (intrkey(K_ESC)) return(APL_CANCEL);
@@ -1725,7 +1707,7 @@ int verbose;
 	if (verbose) {
 		Xlocate(0, L_CMDLINE);
 		Xputterm(L_CLEAR);
-		cp = (dir[0] == '.' && dir[1] == _SC_) ? dir + 2 : dir;
+		cp = (dir[0] == '.' && dir[1] == _SC_) ? &(dir[2]) : dir;
 		Xcprintf2("[%.*k]", n_column - 2, cp);
 		Xtflush();
 	}
@@ -1734,7 +1716,7 @@ int verbose;
 #endif
 
 	if (!funcd1) order = (funcd2) ? ORD_NOPREDIR : ORD_NODIR;
-	strcpy(path, dir);
+	strcpy2(path, dir);
 
 	ret = APL_OK;
 	max = 0;
@@ -1748,8 +1730,9 @@ int verbose;
 	&& (ret = (*funcd1)(dir)) < 0) {
 		if (ret == APL_ERROR) warning(-1, dir);
 		if (dirlist) {
-			for (ndir = 0; ndir < max; ndir++) free(dirlist[ndir]);
-			free(dirlist);
+			for (ndir = 0; ndir < max; ndir++)
+				free2(dirlist[ndir]);
+			free2(dirlist);
 		}
 		return(ret);
 	}
@@ -1769,11 +1752,11 @@ int verbose;
 		else ret = _applydir(dirlist[ndir], funcf,
 			funcd1, funcd2, ORD_NODIR, NULL, verbose);
 		if (ret == APL_CANCEL) {
-			for (; ndir < max; ndir++) free(dirlist[ndir]);
-			free(dirlist);
+			for (; ndir < max; ndir++) free2(dirlist[ndir]);
+			free2(dirlist);
 			return(ret);
 		}
-		free(dirlist[ndir]);
+		free2(dirlist[ndir]);
 
 		if (verbose) {
 			Xlocate(0, L_CMDLINE);
@@ -1782,7 +1765,7 @@ int verbose;
 			Xtflush();
 		}
 	}
-	if (dirlist) free(dirlist);
+	free2(dirlist);
 
 	if (!(dirp = Xopendir(dir))) warning(-1, dir);
 	else {
@@ -1792,7 +1775,7 @@ int verbose;
 				break;
 			}
 			if (isdotdir(dp -> d_name)) continue;
-			strcpy(fname, dp -> d_name);
+			strcpy2(fname, dp -> d_name);
 
 			if (Xlstat(path, &st) < 0) warning(-1, path);
 			else if (s_isdir(&st)) continue;
@@ -1824,12 +1807,12 @@ int verbose;
 }
 
 int applydir(dir, funcf, funcd1, funcd2, order, endmes)
-char *dir;
-int (*funcf)__P_((char *));
-int (*funcd1)__P_((char *));
-int (*funcd2)__P_((char *));
+CONST char *dir;
+int (*funcf)__P_((CONST char *));
+int (*funcd1)__P_((CONST char *));
+int (*funcd2)__P_((CONST char *));
 int order;
-char *endmes;
+CONST char *endmes;
 {
 	char path[MAXPATHLEN];
 	int verbose;
@@ -1843,7 +1826,7 @@ char *endmes;
 		realpath2(dir, path, 0);
 		dir = path;
 	}
-#ifdef	_USEDOSEMU
+#ifdef	DEP_DOSEMU
 	else dir = nodospath(path, dir);
 #endif
 
@@ -1852,7 +1835,7 @@ char *endmes;
 
 /*ARGSUSED*/
 int copyfile(arg, tr)
-char *arg;
+CONST char *arg;
 int tr;
 {
 	int order;
@@ -1864,10 +1847,10 @@ int tr;
 #ifdef	_NOTREE
 	destpath = getdestdir(COPYD_K, arg);
 #else
-# ifdef	_NODOSDRIVE
-	destpath = (tr) ? tree(1, (int *)1) : getdestdir(COPYD_K, arg);
-# else
+# ifdef	DEP_PSEUDOPATH
 	destpath = (tr) ? tree(1, &destdrive) : getdestdir(COPYD_K, arg);
+# else
+	destpath = (tr) ? tree(1, NULL) : getdestdir(COPYD_K, arg);
 # endif
 #endif	/* !_NOTREE */
 
@@ -1876,7 +1859,7 @@ int tr;
 	if (mark > 0
 	|| (!isdir(&(filelist[filepos])) || islink(&(filelist[filepos])))) {
 		if (preparecopy(0, mark) < 0) {
-			free(destpath);
+			free2(destpath);
 			return(FNC_CANCEL);
 		}
 		VOID_C applyfile(safecopy, ENDCP_K);
@@ -1884,13 +1867,13 @@ int tr;
 #ifdef	_NOEXTRACOPY
 	else if (issamedir(destpath, NULL)) {
 		warning(EEXIST, filelist[filepos].name);
-		free(destpath);
+		free2(destpath);
 		return((tr) ? FNC_UPDATE : FNC_CANCEL);
 	}
 #endif
 	else {
 		if (preparecopy(1, 1) < 0) {
-			free(destpath);
+			free2(destpath);
 			return(FNC_CANCEL);
 		}
 		order = (islowerdir()) ? ORD_LOWER : ORD_NORMAL;
@@ -1898,17 +1881,18 @@ int tr;
 			cpdir, touchdir, order, ENDCP_K);
 	}
 
-	free(destpath);
-	if (destdir) free(destdir);
+	free2(destpath);
+	free2(destdir);
 #ifndef	_NOEXTRACOPY
-	if (forwardpath) free(forwardpath);
+	free2(forwardpath);
 	forwardpath = NULL;
 #endif
-#ifndef	_NODOSDRIVE
-	if (destdrive >= 0) shutdrv(destdrive);
+#ifdef	DEP_PSEUDOPATH
+	shutdrv(destdrive);
+	destdrive = -1;
 #endif
-#if	!defined (_NOEXTRACOPY) && !defined (_NODOSDRIVE)
-	if (forwarddrive >= 0) shutdrv(forwarddrive);
+#if	!defined (_NOEXTRACOPY) && defined (DEP_DOSDRIVE)
+	shutdrv(forwarddrive);
 	forwarddrive = -1;
 #endif
 
@@ -1917,14 +1901,11 @@ int tr;
 
 /*ARGSUSED*/
 int movefile(arg, tr)
-char *arg;
+CONST char *arg;
 int tr;
 {
-#ifdef	_USEDOSEMU
+#ifdef	DEP_DOSEMU
 	char path[MAXPATHLEN];
-#endif
-#ifndef	_NOEXTRACOPY
-	int order;
 #endif
 	int ret;
 
@@ -1935,22 +1916,22 @@ int tr;
 #ifdef	_NOTREE
 	destpath = getdestdir(MOVED_K, arg);
 #else
-# ifdef	_NODOSDRIVE
-	destpath = (tr) ? tree(1, (int *)1) : getdestdir(MOVED_K, arg);
-# else
+# ifdef	DEP_PSEUDOPATH
 	destpath = (tr) ? tree(1, &destdrive) : getdestdir(MOVED_K, arg);
+# else
+	destpath = (tr) ? tree(1, NULL) : getdestdir(MOVED_K, arg);
 # endif
 #endif	/* !_NOTREE */
 
 	if (!destpath) return((tr) ? FNC_UPDATE : FNC_CANCEL);
 	if (issamedir(destpath, NULL)) {
-		free(destpath);
+		free2(destpath);
 		return((tr) ? FNC_UPDATE : FNC_CANCEL);
 	}
 	destdir = NULL;
 	if (mark > 0) {
 		if (preparemove(0, mark) < 0) {
-			free(destpath);
+			free2(destpath);
 			return(FNC_CANCEL);
 		}
 		filepos = applyfile(safemove, ENDMV_K);
@@ -1965,21 +1946,20 @@ int tr;
 		if (ret < 0) warning(-1, filelist[filepos].name);
 		else if (ret) {
 			if (preparemove(1, 1) < 0) {
-				free(destpath);
+				free2(destpath);
 				return(FNC_CANCEL);
 			}
-# ifndef	_NODOSDRIVE
+# ifdef	DEP_DOSDRIVE
 			if (dospath3(nullstr)) waitmes();
 # endif
-			order = (islowerdir()) ? ORD_LOWER : ORD_NORMAL;
 			ret = applydir(filelist[filepos].name, safemove,
-				mvdir1, mvdir2, order, ENDMV_K);
+				mvdir1, mvdir2, ORD_NORMAL, ENDMV_K);
 		}
 		else
 #endif	/* !_NOEXTRACOPY */
 		{
 			if (preparemove(0, 1) < 0) {
-				free(destpath);
+				free2(destpath);
 				return(FNC_CANCEL);
 			}
 			ret = safemove(fnodospath(path, filepos));
@@ -1990,17 +1970,18 @@ int tr;
 	}
 
 	if (filepos >= maxfile) filepos = maxfile - 1;
-	free(destpath);
-	if (destdir) free(destdir);
+	free2(destpath);
+	free2(destdir);
 #ifndef	_NOEXTRACOPY
-	if (forwardpath) free(forwardpath);
+	free2(forwardpath);
 	forwardpath = NULL;
 #endif
-#ifndef	_NODOSDRIVE
-	if (destdrive >= 0) shutdrv(destdrive);
+#ifdef	DEP_PSEUDOPATH
+	shutdrv(destdrive);
+	destdrive = -1;
 #endif
-#if	!defined (_NOEXTRACOPY) && !defined (_NODOSDRIVE)
-	if (forwarddrive >= 0) shutdrv(forwarddrive);
+#if	!defined (_NOEXTRACOPY) && defined (DEP_DOSDRIVE)
+	shutdrv(forwarddrive);
 	forwarddrive = -1;
 #endif
 
@@ -2008,7 +1989,7 @@ int tr;
 }
 
 static int forcecpfile(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
@@ -2027,7 +2008,7 @@ char *path;
 }
 
 static int forcecpdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
@@ -2035,7 +2016,7 @@ char *path;
 
 	if (isdotdir(path)) return(APL_OK);
 	if (getdestpath(path, dest, &st) < 0) return(APL_OK);
-	mode = ((st.st_mode & 0777) | S_IWRITE);
+	mode = ((st.st_mode & 0777) | S_IWUSR);
 	if (Xmkdir(dest, mode) >= 0) return(APL_OK);
 	if (errno != EEXIST) return(APL_ERROR);
 	if (stat2(dest, &st) >= 0) {
@@ -2050,7 +2031,7 @@ char *path;
 }
 
 static int forcetouchdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
