@@ -4,39 +4,60 @@
  *	information module
  */
 
-#include <errno.h>
 #include "fd.h"
+#include "dirent.h"
+#include "sysemu.h"
+#include "pathname.h"
+#include "term.h"
+#include "types.h"
 #include "funcno.h"
 #include "kanji.h"
 
-#ifdef	NOERRNO
-extern int errno;
-#endif
-
-#ifdef	USESYSDIRH
-#include <sys/dir.h>
-#endif
-
-#if	MSDOS
-#undef	USEVFSH
-#undef	USEMNTENTH
-#include "unixemu.h"
-#else
-#include <sys/param.h>
-#include <sys/file.h>
-#endif
-
 #ifdef	USEMNTENTH
 #include <mntent.h>
+#endif
+#ifdef	USEMNTTABH
+#include <sys/mnttab.h>
+#endif
+#ifdef	USEMNTCTL
+#include <fshelp.h>
+#include <sys/vfs.h>
+#include <sys/mntctl.h>
+#include <sys/vmount.h>
+#endif
+#if	(!defined (USEMOUNTH) && !defined (USEFSDATA) \
+&& (defined (USEGETFSSTAT) || defined (USEGETVFSTAT) \
+|| defined (USEMNTINFOR) || defined (USEMNTINFO))) \
+|| defined (USEMOUNTH) || defined (USEFSDATA)
+#include <sys/mount.h>
+#endif
+#if	defined (USEGETFSSTAT) || defined (USEGETMNT)
+#include <sys/fs_types.h>
+#endif
+#ifdef	USEGETFSENT
+#include <fstab.h>
+#endif
+#if	defined (USESTATVFSH) || defined (USEGETVFSTAT)
+#include <sys/statvfs.h>
+#endif
+#ifdef	USESTATFSH
+#include <sys/statfs.h>
+#endif
+#ifdef	USEVFSH
+#include <sys/vfs.h>
+#endif
+
+#define	KEYWID			7
+
+#ifdef	USEMNTENTH
 typedef struct mntent		mnt_t;
 #define	setmntent2		setmntent
 #define	getmntent2(f,m)		getmntent(f)
 #define	hasmntopt2(m,o)		strmntopt((m) -> mnt_opts, o)
 #define	endmntent2		endmntent
-#endif	/* USEMNTENTH */
+#endif
 
 #ifdef	USEMNTTABH
-#include <sys/mnttab.h>
 #define	MOUNTED			MNTTAB
 typedef struct mnttab		mnt_t;
 #define	setmntent2		fopen
@@ -51,47 +72,25 @@ typedef struct mnttab		mnt_t;
 # endif
 #endif	/* USEMNTTABH */
 
-#ifdef	USEMNTCTL
-#include <fshelp.h>
-#include <sys/vfs.h>
-#include <sys/mntctl.h>
-#include <sys/vmount.h>
-#endif	/* USEMNTCTL */
-
-#if	!defined (USEMOUNTH) && !defined (USEFSDATA) \
-&& (defined (USEGETFSSTAT) || defined (USEGETVFSTAT) \
-|| defined (USEMNTINFOR) || defined (USEMNTINFO))
-#include <sys/mount.h>
-#endif
-
-#if	defined (USEGETFSSTAT) || defined (USEGETMNT)
-#include <sys/fs_types.h>
-#endif
-
 #if	defined (USEGETFSSTAT) || defined (USEGETVFSTAT) \
 || defined (USEMNTCTL) || defined (USEMNTINFOR) || defined (USEMNTINFO) \
 || defined (USEGETMNT)
 typedef struct _mnt_t {
-	char *mnt_fsname;
-	char *mnt_dir;
-	char *mnt_type;
-	char *mnt_opts;
+	CONST char *mnt_fsname;
+	CONST char *mnt_dir;
+	CONST char *mnt_type;
+	CONST char *mnt_opts;
 } mnt_t;
-static FILE *NEAR setmntent2 __P_((char *, char *));
-static mnt_t *NEAR getmntent2 __P_((FILE *, mnt_t *));
 #define	hasmntopt2(m,o)		strmntopt((m) -> mnt_opts, o)
 # if	defined (USEMNTINFO) || defined (USEGETMNT)
 # define	endmntent2(f)
 # else
-# define	endmntent2	free
+# define	endmntent2	free2
 # endif
-static int mnt_ptr = 0;
-static int mnt_size = 0;
 #endif	/* USEGETFSSTAT || USEGETVFSTAT || USEMNTCTL \
 || USEMNTINFOR || USEMNTINFO || USEGETMNT */
 
 #ifdef	USEGETFSENT
-#include <fstab.h>
 typedef struct fstab		mnt_t;
 #define	setmntent2(f,m)		(FILE *)(setfsent(), NULL)
 #define	getmntent2(f,m)		getfsent()
@@ -100,7 +99,7 @@ typedef struct fstab		mnt_t;
 #define	mnt_dir			fs_file
 #define	mnt_fsname		fs_spec
 #define	mnt_type		fs_vfstype
-#endif	/* USEGETFSENT */
+#endif
 
 #if	MSDOS
 # ifdef	DOUBLESLASH
@@ -109,10 +108,10 @@ typedef struct fstab		mnt_t;
 # define	MNTDIRSIZ	(3 + 1)
 # endif
 typedef struct _mnt_t {
-	char *mnt_fsname;
+	CONST char *mnt_fsname;
 	char mnt_dir[MNTDIRSIZ];
-	char *mnt_type;
-	char *mnt_opts;
+	CONST char *mnt_type;
+	CONST char *mnt_opts;
 } mnt_t;
 #define	hasmntopt2(m,o)		strmntopt((m) -> mnt_opts, o)
 # ifdef	PC98
@@ -129,10 +128,6 @@ typedef struct _mnt_t {
 # endif
 #endif	/* MSDOS */
 
-#if	defined (USESTATVFSH) || defined (USEGETVFSTAT)
-#include <sys/statvfs.h>
-#endif
-
 #ifdef	USESTATVFSH
 # ifdef	USESTATVFS_T
 typedef statvfs_t		statfs_t;
@@ -144,35 +139,31 @@ typedef struct statvfs		statfs_t;
 #endif	/* USESTATVFSH */
 
 #ifdef	USESTATFSH
-#include <sys/statfs.h>
 #define	f_bavail		f_bfree
 typedef struct statfs		statfs_t;
 #define	blocksize(fs)		(fs).f_bsize
-#endif	/* USESTATFSH */
+#endif
 
 #ifdef	USEVFSH
-#include <sys/vfs.h>
 typedef struct statfs		statfs_t;
 #define	blocksize(fs)		(fs).f_bsize
-#endif	/* USEVFSH */
+#endif
 
 #ifdef	USEMOUNTH
-#include <sys/mount.h>
 typedef struct statfs		statfs_t;
 #define	blocksize(fs)		(fs).f_bsize
-#endif	/* USEMOUNTH */
+#endif
 
 #ifdef	USEFSDATA
-#include <sys/mount.h>
 typedef struct fs_data		statfs_t;
-#define	statfs2(path, buf)	(statfs(path, buf) - 1)
+#define	statfs2(p, b)		(statfs(p, b) - 1)
 #define	blocksize(fs)		1024
 #define	f_bsize			fd_req.bsize
 #define	f_files			fd_req.gtot
 #define	f_blocks		fd_req.btot
 #define	f_bfree			fd_req.bfree
 #define	f_bavail		fd_req.bfreen
-#endif	/* USEFSDATA */
+#endif
 
 #ifdef	USEFFSIZE
 #define	f_bsize			f_fsize
@@ -180,15 +171,15 @@ typedef struct fs_data		statfs_t;
 
 #if	defined (USESTATFSH) || defined (USEVFSH) || defined (USEMOUNTH)
 # if	(STATFSARGS >= 4)
-# define	statfs2(path, buf)	statfs(path, buf, sizeof(statfs_t), 0)
-# else
+# define	statfs2(p, b)	statfs(p, b, sizeof(statfs_t), 0)
+# else	/* STATFSARGS < 4 */
 #  if	(STATFSARGS == 3)
-#  define	statfs2(path, buf)	statfs(path, buf, sizeof(statfs_t))
+#  define	statfs2(p, b)	statfs(p, b, sizeof(statfs_t))
 #  else
-#  define	statfs2			statfs
+#  define	statfs2		statfs
 #  endif
-# endif
-#endif
+# endif	/* STATFSARGS < 4 */
+#endif	/* USESTATFSH || USEVFSH || USEMOUNTH */
 
 #if	MSDOS
 typedef struct _statfs_t {
@@ -198,156 +189,169 @@ typedef struct _statfs_t {
 	long f_bavail;
 	long f_files;
 } statfs_t;
-extern int unixstatfs __P_((char *, statfs_t *));
+extern int unixstatfs __P_((CONST char *, statfs_t *));
 #define	statfs2			unixstatfs
 #define	blocksize(fs)		(fs).f_bsize
 #endif
 
-extern VOID error __P_((char *));
-extern int _chdir2 __P_((char *));
-extern char *strcpy2 __P_((char *, char *));
-extern char *strncpy2 __P_((char *, char *, int));
-extern char *getwd2 __P_((VOID_A));
-extern VOID warning __P_((int, char *));
-#ifdef	_USEDOSPATH
-extern int dospath __P_((char *, char *));
-extern char *gendospath __P_((char *, int, int));
+#ifdef	LINUX
+#define	SFN_MSDOSFS
 #endif
-#if	MSDOS && !defined (_NOUSELFN)
-extern int supportLFN __P_((char *));
-# ifndef	_NODOSDRIVE
+#if	defined (FREEBSD) && (__FreeBSD__ < 3)
+# include <osreldate.h>
+# if	defined (__FreeBSD_version) && (__FreeBSD_version < 227000)
+# define	SFN_MSDOSFS
+# endif
+#endif	/* FREEBSD && (__FreeBSD__ < 3) */
+#ifdef	NETBSD
+# if	defined (NetBSD1_0) || defined (NetBSD1_1)
+# define	SFN_MSDOSFS
+# endif
+#endif	/* NETBSD */
+
+#ifndef	MOUNTED
+#define	MOUNTED			"/etc/mtab"
+#endif
+#ifndef	MNTTYPE_43
+#define	MNTTYPE_43		"4.3"		/* NEWS-OS 3-4, HP-UX, HI-UX */
+#endif
+#ifndef	MNTTYPE_42
+#define	MNTTYPE_42		"4.2"		/* SunOS 4 */
+#endif
+#ifndef	MNTTYPE_UFS
+#define	MNTTYPE_UFS		"ufs"		/* SVR4, OSF/1, Free/NetBSD */
+#endif
+#ifndef	MNTTYPE_FFS
+#define	MNTTYPE_FFS		"ffs"		/* NetBSD, OpenBSD */
+#endif
+#ifndef	MNTTYPE_ADVFS
+#define	MNTTYPE_ADVFS		"advfs"		/* OSF/1 */
+#endif
+#ifndef	MNTTYPE_VXFS
+#define	MNTTYPE_VXFS		"vxfs"		/* HP-UX */
+#endif
+#ifndef	MNTTYPE_HFS
+#define	MNTTYPE_HFS		"hfs"		/* Darwin, HP-UX */
+#endif
+#ifndef	MNTTYPE_EXT2
+#define	MNTTYPE_EXT2		"ext2"		/* Linux */
+#endif
+#ifndef	MNTTYPE_EXT3
+#define	MNTTYPE_EXT3		"ext3"		/* Linux */
+#endif
+#ifndef	MNTTYPE_JFS
+#define	MNTTYPE_JFS		"jfs"		/* AIX */
+#endif
+#ifndef	MNTTYPE_EFS
+#define	MNTTYPE_EFS		"efs"		/* IRIX */
+#endif
+#ifndef	MNTTYPE_SYSV
+#define	MNTTYPE_SYSV		"sysv"		/* SystemV Rel.3 */
+#endif
+#ifndef	MNTTYPE_DGUX
+#define	MNTTYPE_DGUX		"dg/ux"		/* DG/UX */
+#endif
+#ifndef	MNTTYPE_UMSDOS
+#define	MNTTYPE_UMSDOS		"umsdos"	/* umsdosfs on Linux */
+#endif
+#ifndef	MNTTYPE_MSDOS
+#define	MNTTYPE_MSDOS		"msdos"		/* msdosfs */
+#endif
+#ifndef	MNTTYPE_MSDOSFS
+#define	MNTTYPE_MSDOSFS		"msdosfs"	/* msdosfs on FreeBSD */
+#endif
+#ifndef	MNTTYPE_VFAT
+#define	MNTTYPE_VFAT		"vfat"		/* vfatfs on Linux */
+#endif
+#ifndef	MNTTYPE_PC
+#define	MNTTYPE_PC		"pc"		/* MS-DOS */
+#endif
+#ifndef	MNTTYPE_DOS7
+#define	MNTTYPE_DOS7		"dos7"		/* MS-DOS on Win95 */
+#endif
+#ifndef	MNTTYPE_FAT12
+#define	MNTTYPE_FAT12		"fat12"		/* MS-DOS */
+#endif
+#ifndef	MNTTYPE_FAT16
+#define	MNTTYPE_FAT16		"fat16(16bit sector)"	/* MS-DOS */
+#endif
+#ifndef	MNTTYPE_FAT16X
+#define	MNTTYPE_FAT16X		"fat16(32bit sector)"	/* MS-DOS */
+#endif
+#ifndef	MNTTYPE_FAT32
+#define	MNTTYPE_FAT32		"fat32"		/* Win98 */
+#endif
+#ifndef	MNTTYPE_SHARED
+#define	MNTTYPE_SHARED		"shared"	/* Win98 */
+#endif
+#define	MNTTYPE_XNFS		"nfs"		/* NFS */
+
+extern VOID error __P_((CONST char *));
+extern int _chdir2 __P_((CONST char *));
+extern VOID warning __P_((int, CONST char *));
+#ifdef	DEP_DOSLFN
+extern int supportLFN __P_((CONST char *));
+# ifdef	DEP_DOSDRIVE
 extern int checkdrive __P_((int));
 # endif
-#endif	/* MSDOS && !_NOUSELFN */
-#ifndef	_NODOSDRIVE
+#endif	/* DEP_DOSLFN */
+#ifdef	DEP_DOSDRIVE
 extern int dosstatfs __P_((int, char *));
 #endif
-extern char *malloc2 __P_((ALLOC_T));
-extern char *realloc2 __P_((VOID_P, ALLOC_T));
-#if	!MSDOS || !defined (NOFLOCK)
-extern int Xstat __P_((char *, struct stat *));
-#endif
-extern int Xaccess __P_((char *, int));
 extern int filetop __P_((int));
 extern VOID cputspace __P_((int));
-extern VOID cputstr __P_((int, char *));
-#ifdef	_NOPTY
+extern VOID cputstr __P_((int, CONST char *));
+#ifdef	DEP_PTY
+extern VOID Xlocate __P_((int, int));
+extern VOID Xputterm __P_((int));
+extern VOID Xputch2 __P_((int));
+extern VOID Xcputs2 __P_((CONST char *));
+extern VOID Xcprintf2 __P_((CONST char *, ...));
+#else
 #define	Xlocate			locate
 #define	Xputterm		putterm
 #define	Xputch2			putch2
 #define	Xcputs2			cputs2
 #define	Xcprintf2		cprintf2
-#else
-extern VOID Xlocate __P_((int, int));
-extern VOID Xputterm __P_((int));
-extern VOID Xputch2 __P_((int));
-extern VOID Xcputs2 __P_((char *));
-extern VOID Xcprintf2 __P_((CONST char *, ...));
 #endif
 
-extern bindtable bindlist[];
-extern functable funclist[];
+extern bindlist_t bindlist;
+extern int maxbind;
+extern CONST functable funclist[];
 extern char fullpath[];
 extern char *distributor;
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_DOSDRIVE
 extern int needbavail;
 #endif
 
-#define	KEYWID		7
-
-#ifndef	MOUNTED
-#define	MOUNTED		"/etc/mtab"
-#endif
-
-#ifndef	MNTTYPE_43
-#define	MNTTYPE_43	"4.3"	/* NEWS-OS 3-4, HP-UX, HI-UX */
-#endif
-#ifndef	MNTTYPE_42
-#define	MNTTYPE_42	"4.2"	/* SunOS 4 */
-#endif
-#ifndef	MNTTYPE_UFS
-#define	MNTTYPE_UFS	"ufs"	/* SVR4, OSF/1, FreeBSD, NetBSD */
-#endif
-#ifndef	MNTTYPE_FFS
-#define	MNTTYPE_FFS	"ffs"	/* NetBSD, OpenBSD */
-#endif
-#ifndef	MNTTYPE_ADVFS
-#define	MNTTYPE_ADVFS	"advfs"	/* OSF/1 */
-#endif
-#ifndef	MNTTYPE_VXFS
-#define	MNTTYPE_VXFS	"vxfs"	/* HP-UX */
-#endif
-#ifndef	MNTTYPE_HFS
-#define	MNTTYPE_HFS	"hfs"	/* Darwin, HP-UX */
-#endif
-#ifndef	MNTTYPE_EXT2
-#define	MNTTYPE_EXT2	"ext2"	/* Linux */
-#endif
-#ifndef	MNTTYPE_EXT3
-#define	MNTTYPE_EXT3	"ext3"	/* Linux */
-#endif
-#ifndef	MNTTYPE_JFS
-#define	MNTTYPE_JFS	"jfs"	/* AIX */
-#endif
-#ifndef	MNTTYPE_EFS
-#define	MNTTYPE_EFS	"efs"	/* IRIX */
-#endif
-#ifndef	MNTTYPE_SYSV
-#define	MNTTYPE_SYSV	"sysv"	/* SystemV Rel.3 */
-#endif
-#ifndef	MNTTYPE_DGUX
-#define	MNTTYPE_DGUX	"dg/ux"	/* DG/UX */
-#endif
-#ifndef	MNTTYPE_MSDOS
-#define	MNTTYPE_MSDOS	"msdos"	/* msdosfs */
-#endif
-#ifndef	MNTTYPE_UMSDOS
-#define	MNTTYPE_UMSDOS	"umsdos"	/* umsdosfs */
-#endif
-#ifndef	MNTTYPE_VFAT
-#define	MNTTYPE_VFAT	"vfat"	/* vfatfs */
-#endif
-#ifndef	MNTTYPE_PC
-#define	MNTTYPE_PC	"pc"	/* MS-DOS */
-#endif
-#ifndef	MNTTYPE_DOS7
-#define	MNTTYPE_DOS7	"dos7"	/* MS-DOS on Win95 */
-#endif
-#ifndef	MNTTYPE_FAT12
-#define	MNTTYPE_FAT12	"fat12"	/* MS-DOS */
-#endif
-#ifndef	MNTTYPE_FAT16
-#define	MNTTYPE_FAT16	"fat16(16bit sector)"	/* MS-DOS */
-#endif
-#ifndef	MNTTYPE_FAT16X
-#define	MNTTYPE_FAT16X	"fat16(32bit sector)"	/* MS-DOS */
-#endif
-#ifndef	MNTTYPE_FAT32
-#define	MNTTYPE_FAT32	"fat32"	/* Win98 */
-#endif
-#ifndef	MNTTYPE_SHARED
-#define	MNTTYPE_SHARED	"shared"	/* Win98 */
-#endif
-#define	MNTTYPE_XNFS	"nfs"	/* NFS */
-
 static int NEAR code2str __P_((char *, int));
 static int NEAR checkline __P_((int));
 VOID help __P_((int));
-static int NEAR getfsinfo __P_((char *, statfs_t *, mnt_t *));
-static char *NEAR strmntopt __P_((char *, char *));
-int writablefs __P_((char *));
-off_t getblocksize __P_((char *));
-static int NEAR info1line __P_((int, int, char *, off_t, char *, char *));
+#if	defined (USEGETFSSTAT) || defined (USEGETVFSTAT) \
+|| defined (USEMNTCTL) || defined (USEMNTINFOR) || defined (USEMNTINFO) \
+|| defined (USEGETMNT)
+static FILE *NEAR setmntent2 __P_((CONST char *, CONST char *));
+static mnt_t *NEAR getmntent2 __P_((FILE *, mnt_t *));
+#endif
+static int NEAR getfsinfo __P_((CONST char *, statfs_t *, mnt_t *));
+static CONST char *NEAR strmntopt __P_((CONST char *, CONST char *));
+#ifndef	NOFLOCK
+int isnfs __P_((CONST char *));
+#endif
+int writablefs __P_((CONST char *));
+off_t getblocksize __P_((CONST char *));
+static int NEAR info1line __P_((int, int, CONST char *,
+		off_t, CONST char *, CONST char *));
 off_t calcKB __P_((off_t, off_t));
-int getinfofs __P_((char *, off_t *, off_t *, off_t *));
-int infofs __P_((char *));
+int getinfofs __P_((CONST char *, off_t *, off_t *, off_t *));
+int infofs __P_((CONST char *));
 
 static CONST int keycodelist[] = {
 	K_HOME, K_END, K_DL, K_IL, K_DC, K_IC,
 	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER, K_HELP,
 	K_BS, '\t', K_CR, K_ESC
 };
-#define	KEYCODESIZ	arraysize(keycodelist)
+#define	KEYCODESIZ		arraysize(keycodelist)
 static CONST char *keystrlist[] = {
 	"Home", "End", "DelLin", "InsLin", "Del", "Ins",
 	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter", "Help",
@@ -359,7 +363,7 @@ static CONST strtable mntlist[] = {
 	{FSID_LFN, MNTTYPE_FAT32},
 	{FSID_LFN, MNTTYPE_SHARED},
 #if	MSDOS
-# ifndef	_NODOSDRIVE
+# ifdef	DEP_DOSDRIVE
 	{FSID_FAT, MNTTYPE_FAT12},
 	{FSID_FAT, MNTTYPE_FAT16},
 	{FSID_FAT, MNTTYPE_FAT16X},
@@ -375,12 +379,13 @@ static CONST strtable mntlist[] = {
 	{FSID_SYSV, MNTTYPE_DGUX},
 	{FSID_LINUX, MNTTYPE_EXT2},
 	{FSID_LINUX, MNTTYPE_EXT3},
-# ifdef	LINUX
+	{FSID_FAT, MNTTYPE_UMSDOS},
+# ifdef	SFN_MSDOSFS
 	{FSID_FAT, MNTTYPE_MSDOS},
 # else
 	{FSID_LFN, MNTTYPE_MSDOS},
 # endif
-	{FSID_LFN, MNTTYPE_UMSDOS},
+	{FSID_LFN, MNTTYPE_MSDOSFS},
 	{FSID_LFN, MNTTYPE_VFAT},
 	{0, MNTTYPE_ADVFS},
 	{0, MNTTYPE_VXFS},
@@ -390,13 +395,21 @@ static CONST strtable mntlist[] = {
 # endif
 #endif	/* !MSDOS */
 };
-#define	MNTLISTSIZ	arraysize(mntlist)
+#define	MNTLISTSIZ		arraysize(mntlist)
+#if	defined (USEGETFSSTAT) || defined (USEGETVFSTAT) \
+|| defined (USEMNTCTL) || defined (USEMNTINFOR) || defined (USEMNTINFO) \
+|| defined (USEGETMNT)
+static int mnt_ptr = 0;
+static int mnt_size = 0;
+#endif
 
 
 static int NEAR code2str(buf, code)
 char *buf;
 int code;
 {
+	int i;
+
 	buf = buf + strlen(buf);
 	if (code >= K_F(1) && code <= K_F(20))
 		snprintf2(buf, KEYWID + 1, "F%-6d", code - K_F0);
@@ -411,8 +424,6 @@ int code;
 	else if (code == K_LEFT)
 		snprintf2(buf, KEYWID + 1, "%-*.*s", KEYWID, KEYWID, LEFAR_K);
 	else {
-		int i;
-
 		for (i = 0; i < KEYCODESIZ; i++)
 			if (code == keycodelist[i]) break;
 		if (i < KEYCODESIZ)
@@ -453,6 +464,7 @@ int y;
 VOID help(arch)
 int arch;
 {
+	CONST char *cp;
 	char buf[(KEYWID + 1 + 1) * 2 + 1];
 	int i, j, c, x, y, yy;
 
@@ -480,23 +492,23 @@ int arch;
 
 		c = 0;
 		buf[0] = '\0';
-		for (j = 0; j < MAXBINDTABLE && bindlist[j].key >= 0; j++)
-			if (i == (int)(bindlist[j].f_func)) {
-				c += code2str(buf, (int)(bindlist[j].key));
-				if (c >= 2) break;
-			}
-		if (c < 2)
-		for (j = 0; j < MAXBINDTABLE && bindlist[j].key >= 0; j++)
-			if (i == (int)(bindlist[j].d_func)) {
-				c += code2str(buf, (int)(bindlist[j].key));
-				if (c >= 2) break;
-			}
+		for (j = 0; j < maxbind; j++) {
+			if (i != ffunc(j)) continue;
+			c += code2str(buf, (int)(bindlist[j].key));
+			if (c >= 2) break;
+		}
+		if (c < 2) for (j = 0; j < maxbind; j++) {
+			if (i != dfunc(j)) continue;
+			c += code2str(buf, (int)(bindlist[j].key));
+			if (c >= 2) break;
+		}
 		if (!c) continue;
 
 		Xcputs2("  ");
 		if (c < 2) cputspace(KEYWID);
-		Xcprintf2("%k: %k",
-			buf, mesconv(funclist[i].hmes, funclist[i].hmes_eng));
+		cp = mesconv2(funclist[i].hmes_no,
+			funclist[i].hmes, funclist[i].hmes_eng);
+		Xcprintf2("%k: %k", buf, cp);
 
 		y = checkline(y);
 	}
@@ -513,7 +525,7 @@ int arch;
 #ifdef	USEMNTCTL
 /*ARGSUSED*/
 static FILE *NEAR setmntent2(file, mode)
-char *file, *mode;
+CONST char *file, *mode;
 {
 	char *buf;
 
@@ -552,7 +564,7 @@ mnt_t *mntp;
 			+ vmntp -> vmt_data[VMT_HOSTNAME].vmt_off]);
 		len += strlen(host) + 1;
 		fsname = realloc2(fsname, len);
-		strcpy(strcpy2(strcpy2(fsname, host), ":"), cp);
+		strcpy2(strcpy2(strcpy2(fsname, host), ":"), cp);
 	}
 
 	cp = &(buf[mnt_ptr + vmntp -> vmt_data[VMT_STUB].vmt_off]);
@@ -568,7 +580,7 @@ mnt_t *mntp;
 		memcpy(type, cp, len);
 	}
 	else if (type) {
-		free(type);
+		free2(type);
 		type = NULL;
 	}
 
@@ -601,7 +613,7 @@ typedef struct statfs		mntinfo_t;
 
 /*ARGSUSED*/
 static FILE *NEAR setmntent2(file, mode)
-char *file, *mode;
+CONST char *file, *mode;
 {
 # ifndef	USEMNTINFO
 	int size;
@@ -646,25 +658,30 @@ mnt_t *mntp;
 # if	defined (USEMNTINFO) || defined (USEGETVFSTAT)
 #  ifdef	USEVFCNAME
 	struct vfsconf *conf;
-#  define	getvfsbynumber(n)	((conf = getvfsbytype(n)) \
-					? conf -> vfc_name : NULL)
+#  define	getvfsbynumber(n) \
+				((conf = getvfsbytype(n)) \
+				? conf -> vfc_name : NULL)
 #  else	/* !USEVFCNAME */
 #   ifdef	USEFFSTYPE
-#   define	getvfsbynumber(n)	(buf[mnt_ptr].f_fstypename)
+#   define	getvfsbynumber(n) \
+				(buf[mnt_ptr].f_fstypename)
 #   else	/* !USEFFSTYPE */
 #    ifdef	INITMOUNTNAMES
 	static char *mnt_names[] = INITMOUNTNAMES;
-#    define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
-					? mnt_names[n] : NULL)
+#    define	getvfsbynumber(n) \
+				(((n) <= MOUNT_MAXTYPE) \
+				? mnt_names[n] : NULL)
 #    else
-#    define	getvfsbynumber(n)	(NULL)
+#    define	getvfsbynumber(n) \
+				(NULL)
 #    endif
 #   endif	/* !USEFFSTYPE */
 #  endif	/* !USEVFCNAME */
 # else	/* !USEMNTINFO && !USEGETVFSTAT */
 #  ifdef	USEGETFSSTAT
-#  define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
-					? mnt_names[n] : NULL)
+#  define	getvfsbynumber(n) \
+				(((n) <= MOUNT_MAXTYPE) \
+				? mnt_names[n] : NULL)
 #  endif
 # endif	/* !USEMNTINFO && !USEGETVFSTAT */
 	static char *fsname = NULL;
@@ -695,7 +712,7 @@ mnt_t *mntp;
 		memcpy(type, cp, len);
 	}
 	else if (type) {
-		free(type);
+		free2(type);
 		type = NULL;
 	}
 # ifdef	DEBUG
@@ -720,7 +737,7 @@ mnt_t *mntp;
 #ifdef	USEGETMNT
 /*ARGSUSED*/
 static FILE *NEAR setmntent2(file, mode)
-char *file, *mode;
+CONST char *file, *mode;
 {
 	mnt_ptr = 0;
 
@@ -763,23 +780,35 @@ mnt_t *mntp;
 #endif	/* USEGETMNT */
 
 static int NEAR getfsinfo(path, fsbuf, mntbuf)
-char *path;
+CONST char *path;
 statfs_t *fsbuf;
 mnt_t *mntbuf;
 {
 #if	MSDOS
-# ifndef	_NODOSDRIVE
+# ifdef	DEP_DOSDRIVE
 	int i;
 # endif
-# ifdef	DOUBLESLASH
-	char *cp;
-	int len;
-# endif
-	statfs_t fs;
-	mnt_t mnt;
 	char buf[MAXPATHLEN];
 	int drive;
+#else	/* !MSDOS */
+# ifdef	DEP_DOSDRIVE
+	static char dosmntdir[4];
+	char buf[3 * sizeof(long) + 1];
+	int drv;
+# endif
+	mnt_t *mntp;
+	FILE *fp;
+	char fsname[MAXPATHLEN], dir[MAXPATHLEN];
+	ALLOC_T match;
+#endif	/* !MSDOS */
+#if	!MSDOS || defined (DOUBLESLASH)
+	CONST char *cp;
+	ALLOC_T len;
+#endif
+	statfs_t fs;
+	mnt_t mnt;
 
+#if	MSDOS
 	if (!fsbuf) fsbuf = &fs;
 	if (!mntbuf) mntbuf = &mnt;
 
@@ -801,9 +830,7 @@ mnt_t *mntbuf;
 # endif
 	VOID_C gendospath(mntbuf -> mnt_dir, drive, _SC_);
 
-# ifdef	_NOUSELFN
-	mntbuf -> mnt_type = MNTTYPE_PC;
-# else	/* !_NOUSELFN */
+# ifdef	DEP_DOSLFN
 	switch (supportLFN(mntbuf -> mnt_dir)) {
 		case 3:
 			mntbuf -> mnt_type = MNTTYPE_SHARED;
@@ -814,7 +841,7 @@ mnt_t *mntbuf;
 		case 1:
 			mntbuf -> mnt_type = MNTTYPE_DOS7;
 			break;
-#  ifndef	_NODOSDRIVE
+#  ifdef	DEP_DOSDRIVE
 		case -1:
 			mntbuf -> mnt_fsname = "LFN emurate";
 			mntbuf -> mnt_type = MNTTYPE_DOS7;
@@ -831,48 +858,37 @@ mnt_t *mntbuf;
 			break;
 #  endif
 		case -4:
-			errno = ENOENT;
-			return(-1);
+			return(seterrno(ENOENT));
 /*NOTREACHED*/
 			break;
 		default:
 			mntbuf -> mnt_type = MNTTYPE_PC;
 			break;
 	}
-# endif	/* !_NOUSELFN */
+# else	/* !DEP_DOSLFN */
+	mntbuf -> mnt_type = MNTTYPE_PC;
+# endif	/* !DEP_DOSLFN */
 	mntbuf -> mnt_opts = nullstr;
 
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0) return(-1);
 #else	/* !MSDOS */
-# ifndef	_NODOSDRIVE
-	int drv;
-# endif
-	statfs_t fs;
-	mnt_t mnt;
-	mnt_t *mntp;
-	FILE *fp;
-	char *dir, fsname[MAXPATHLEN];
-	ALLOC_T len, match;
-
 	if (!fsbuf) fsbuf = &fs;
 	if (!mntbuf) mntbuf = &mnt;
 
-	dir = NULL;
+	cp = NULL;
 	if (!strncmp(path, "/dev/", strsize("/dev/"))) {
-		if (_chdir2(path) < 0) dir = path;
+		cp = getrealpath(path, dir, NULL);
+		if (cp != dir && cp != path) return(-1);
 	}
-# ifndef	_NODOSDRIVE
+# ifdef	DEP_DOSDRIVE
 	else if ((drv = dospath(path, NULL))) {
-		static char dosmntdir[4];
-		char buf[3 * sizeof(long) + 1];
-
-		mntbuf -> mnt_fsname = nullstr;
+		mntbuf -> mnt_fsname = vnullstr;
 		mntbuf -> mnt_dir = dosmntdir;
 		dosmntdir[0] = drv;
-		strcpy(&(dosmntdir[1]), ":\\");
+		strcpy2(&(dosmntdir[1]), ":\\");
 		mntbuf -> mnt_type =
 			(islower2(drv)) ? MNTTYPE_DOS7 : MNTTYPE_PC;
-		mntbuf -> mnt_opts = nullstr;
+		mntbuf -> mnt_opts = vnullstr;
 		if (dosstatfs(drv, buf) < 0) return(-1);
 		if (buf[3 * sizeof(long)] & 001)
 			mntbuf -> mnt_type = MNTTYPE_FAT32;
@@ -895,12 +911,10 @@ mnt_t *mntbuf;
 
 		return(0);
 	}
-# endif	/* !_NODOSDRIVE */
-	else if (_chdir2(path) < 0) return(-1);
+# endif	/* DEP_DOSDRIVE */
+	else if ((cp = getrealpath(path, dir, NULL)) != dir) return(-1);
 
-	if (!dir) {
-		dir = getwd2();
-		if (_chdir2(fullpath) < 0) error(fullpath);
+	if (cp == dir) {
 		match = (ALLOC_T)0;
 
 		if (!(fp = setmntent2(MOUNTED, "rb"))) return(-1);
@@ -923,47 +937,39 @@ mnt_t *mntbuf;
 			&& dir[len] && dir[len] != _SC_))
 				continue;
 			match = len;
-			strcpy(fsname, mntp -> mnt_fsname);
+			strcpy2(fsname, mntp -> mnt_fsname);
 		}
 		endmntent2(fp);
 
-		free(dir);
-		if (!match) {
-			errno = ENOENT;
-			return(-1);
-		}
-		dir = fsname;
+		if (!match) return(seterrno(ENOENT));
+		cp = fsname;
 	}
 
 	if (!(fp = setmntent2(MOUNTED, "rb"))) return(-1);
 	while ((mntp = getmntent2(fp, &mnt)))
-		if (!strcmp(dir, mntp -> mnt_fsname)) break;
+		if (!strcmp(cp, mntp -> mnt_fsname)) break;
 	endmntent2(fp);
-	if (!mntp) {
-		errno = ENOENT;
-		return(-1);
-	}
+	if (!mntp) return(seterrno(ENOENT));
 	memcpy((char *)mntbuf, (char *)mntp, sizeof(mnt_t));
 
-	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0
-	&& (path == dir || statfs2(path, fsbuf) < 0))
-		return(-1);
+	if (statfs2(mntbuf -> mnt_dir, fsbuf) >= 0) /*EMPTY*/;
+	else if (path == cp || statfs2(path, fsbuf) < 0) return(-1);
 #endif	/* !MSDOS */
 
 	return(0);
 }
 
-static char *NEAR strmntopt(s1, s2)
-char *s1, *s2;
+static CONST char *NEAR strmntopt(s1, s2)
+CONST char *s1, *s2;
 {
-	char *cp;
+	CONST char *cp;
 	ALLOC_T len;
 
 	len = strlen(s2);
 	for (cp = s1; cp && *cp;) {
 		if (!strncmp(cp, s2, len) && (!cp[len] || cp[len] == ','))
 			return(cp);
-		if ((cp = strchr(cp, ','))) cp++;
+		if ((cp = strchr2(cp, ','))) cp++;
 	}
 
 	return(NULL);
@@ -971,7 +977,7 @@ char *s1, *s2;
 
 #ifndef	NOFLOCK
 int isnfs(path)
-char *path;
+CONST char *path;
 {
 	mnt_t mntbuf;
 	struct stat st;
@@ -992,22 +998,21 @@ char *path;
 #endif	/* NOFLOCK */
 
 int writablefs(path)
-char *path;
+CONST char *path;
 {
-#ifdef	_USEDOSEMU
+#ifdef	DEP_DOSEMU
 	int drv;
 #endif
 	mnt_t mntbuf;
 	int i;
 
 	if (Xaccess(path, R_OK | W_OK | X_OK) < 0) return(-1);
-#ifdef	_USEDOSEMU
+#ifdef	DEP_DOSEMU
 	if ((drv = dospath(path, NULL)))
 		return((isupper2(drv)) ? FSID_FAT : FSID_DOSDRIVE);
 #endif
-	if (getfsinfo(path, NULL, &mntbuf) < 0
-	|| hasmntopt2(&mntbuf, "ro"))
-		return(0);
+	if (getfsinfo(path, NULL, &mntbuf) < 0) return(0);
+	if (hasmntopt2(&mntbuf, "ro")) return(0);
 
 	for (i = 0; i < MNTLISTSIZ; i++)
 		if (!strcmp(mntbuf.mnt_type, mntlist[i].str))
@@ -1018,7 +1023,7 @@ char *path;
 
 /*ARGSUSED*/
 off_t getblocksize(dir)
-char *dir;
+CONST char *dir;
 {
 #if	MSDOS
 	statfs_t fsbuf;
@@ -1044,19 +1049,15 @@ char *dir;
 		if (Xstat(dir, &st) >= 0) return((off_t)(st.st_blksize));
 	}
 
-# ifdef	DEV_BSIZE
 	return(DEV_BSIZE);
-# else
-	return((off_t)512);
-# endif
 #endif	/* !MSDOS */
 }
 
 static int NEAR info1line(yy, y, ind, n, s, unit)
 int yy, y;
-char *ind;
+CONST char *ind;
 off_t n;
-char *s, *unit;
+CONST char *s, *unit;
 {
 	int width;
 
@@ -1090,17 +1091,17 @@ off_t block, byte;
 }
 
 int getinfofs(path, totalp, freep, bsizep)
-char *path;
+CONST char *path;
 off_t *totalp, *freep, *bsizep;
 {
 	statfs_t fsbuf;
 	int n;
 
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_DOSDRIVE
 	needbavail++;
 #endif
 	n = getfsinfo(path, &fsbuf, NULL);
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_DOSDRIVE
 	needbavail--;
 #endif
 	if (n < 0) {
@@ -1116,18 +1117,18 @@ off_t *totalp, *freep, *bsizep;
 }
 
 int infofs(path)
-char *path;
+CONST char *path;
 {
 	statfs_t fsbuf;
 	mnt_t mntbuf;
 	int y, yy;
 
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_DOSDRIVE
 	needbavail++;
 #endif
 	if (getfsinfo(path, &fsbuf, &mntbuf) < 0) {
 		warning(ENOTDIR, path);
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_DOSDRIVE
 		needbavail--;
 #endif
 		return(0);
@@ -1159,7 +1160,7 @@ char *path;
 		}
 		warning(0, HITKY_K);
 	}
-#ifndef	_NODOSDRIVE
+#ifdef	DEP_DOSDRIVE
 	needbavail--;
 #endif
 
